"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from .users import SpecificUser, Broadcaster
from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from typing import List, Optional, Literal


# Raids
class RaidInfo(TypedDict):
    """
    Represents information about a raid.

    Attributes
    ----------
    created_at: str
        The timestamp when the raid was created.
    is_mature: bool
        Indicates if the raid is flagged as mature.
    """
    created_at: str
    is_mature: bool


# Streams
class StreamKey(TypedDict):
    """
    Represents a stream key.

    Attributes
    ----------
    stream_key: str
        The stream key used to start a stream.
    """
    stream_key: str


class StreamInfo(SpecificUser):
    """
    Represents a stream.

    Attributes
    ----------
    id: str
        The unique identifier for the stream.
    game_id: str
        The ID of the game being streamed.
    game_name: str
        The name of the game being streamed.
    type: str
        The type of stream (live, etc.).
    title: str
        The title of the stream.
    tags: List[str]
        A list of tags associated with the stream.
    viewer_count: int
        The number of viewers watching the stream.
    started_at: str
        The timestamp when the stream started.
    language: str
        The language of the stream.
    thumbnail_url: str
        The URL for the stream's thumbnail image.
    is_mature: bool
        Indicates if the stream is marked as mature content.
    """
    id: str
    game_id: str
    game_name: str
    type: str
    title: str
    tags: List[str]
    viewer_count: int
    started_at: str
    language: str
    thumbnail_url: str
    is_mature: bool


class StreamMarkerInfo(TypedDict):
    """
    Represents information about a stream marker.

    Attributes
    ----------
    id: str
        The unique identifier for the stream marker.
    created_at: str
        The timestamp when the marker was created.
    position_seconds: int
        The position in seconds where the marker was placed.
    description: str
        A description of the marker.
    """
    id: str
    created_at: str
    position_seconds: int
    description: str


class VideoMarker(TypedDict):
    """
    Represents a marker in a video.

    Attributes
    ----------
    id: str
        The unique identifier for the marker.
    created_at: str
        The timestamp when the marker was created.
    description: str
        A description of the marker.
    position_seconds: int
        The position in seconds where the marker was placed.
    url: str
        The URL to access the marker.
    """
    id: str
    created_at: str
    description: str
    position_seconds: int
    url: str


class Video(TypedDict):
    """
    Represents a video with markers.

    Attributes
    ----------
    video_id: str
        The unique identifier for the video.
    markers: List[Marker]
        A list of markers associated with the video.
    """
    video_id: str
    markers: List[VideoMarker]


class StreamMarker(SpecificUser):
    """
    Represents a stream marker.

    Attributes
    ----------
    videos: List[Video]
        A list of videos with associated markers.
    """
    videos: List[Video]


# Ads
class CommercialStatus(TypedDict):
    """
    Represents the status of a commercial.

    Attributes
    ----------
    length: int
        The length of the commercial in seconds.
    message: str
        A message associated with the commercial status.
    retry_after: int
        The time in seconds to wait before retrying the commercial.
    """
    length: int
    message: str
    retry_after: int


class AdSnooze(TypedDict):
    """
    Represents the snooze status of ads.

    Attributes
    ----------
    snooze_count: int
        The number of times ads have been snoozed.
    snooze_refresh_at: str
        The timestamp when the snooze refreshes.
    next_ad_at: str
        The timestamp when the next ad is scheduled.
    """
    snooze_count: int
    snooze_refresh_at: str
    next_ad_at: str


class AdSchedule(AdSnooze):
    """
    Represents an ad schedule.

    Attributes
    ----------
    duration: int
        The duration of the ad schedule.
    last_ad_at: str
        The timestamp of the last ad.
    preroll_free_time: int
        The time in seconds before preroll ads are shown.
    """
    duration: int
    last_ad_at: str
    preroll_free_time: int


# Schedule
class ScheduleCategory(TypedDict):
    """
    Represents a schedule category.

    Attributes
    ----------
    id: str
        The unique identifier for the category.
    name: str
        The name of the category.
    """
    id: str
    name: str


class ScheduleSegment(TypedDict):
    """
    Represents a segment in a schedule.

    Attributes
    ----------
    id: str
        The unique identifier for the segment.
    start_time: str
        The start time of the segment.
    end_time: str
        The end time of the segment.
    title: str
        The title of the segment.
    canceled_until: Optional[str]
        The time until which the segment is canceled, if applicable.
    category: Optional[ScheduleCategory]
        The category associated with the segment.
    is_recurring: bool
        Indicates if the segment is recurring.
    """
    id: str
    start_time: str
    end_time: str
    title: str
    canceled_until: Optional[str]
    category: Optional[ScheduleCategory]
    is_recurring: bool


class ScheduleVacation(TypedDict):
    """
    Represents a vacation in a schedule.

    Attributes
    ----------
    start_time: str
        The start time of the vacation.
    end_time: str
        The end time of the vacation.
    """
    start_time: str
    end_time: str


class Schedule(Broadcaster):
    """
    Represents a broadcaster's schedule.

    Attributes
    ----------
    segments: List[ScheduleSegment]
        A list of segments in the schedule.
    vacation: Optional[ScheduleVacation]
        The vacation period in the schedule, if applicable.
    """
    segments: List[ScheduleSegment]
    vacation: Optional[ScheduleVacation]


# CCLs
Locale = Literal[
    'bg-BG',
    'cs-CZ',
    'da-DK',
    'de-DE',
    'el-GR',
    'en-GB',
    'en-US',
    'es-ES',
    'es-MX',
    'fi-FI',
    'fr-FR',
    'hu-HU',
    'it-IT',
    'ja-JP',
    'ko-KR',
    'nl-NL',
    'no-NO',
    'pl-PL',
    'pt-BT',
    'pt-PT',
    'ro-RO',
    'ru-RU',
    'sk-SK',
    'sv-SE',
    'th-TH',
    'tr-TR',
    'vi-VN',
    'zh-CN',
    'zh-TW']


class CCLInfo(TypedDict):
    """
    Represents content classification labels (CCLs) information.

    Attributes
    ----------
    id: str
        The unique identifier for the CCL.
    description: str
        The description of the CCL.
    name: str
        The name of the CCL.
    """
    id: str
    description: str
    name: str



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Optional, Literal, List, Dict, Any, AsyncGenerator, Tuple
    from .types import Data, TData, Edata, chat, users, bits, moderation, PData
    from .state import ConnectionState
    from .user import User

__all__ = ('Chat', 'BroadcasterChat')


class Chat:
    """
    Represents a channel chat.
    """
    __slots__ = ('_state', '_user_id', '_auth_user_id')

    def __init__(self, user_id: str, auth_user_id: str, *, state: ConnectionState) -> None:
        self._state: ConnectionState = state
        self._user_id: str = user_id
        self._auth_user_id: str = auth_user_id

    async def get_settings(self) -> chat.Settings:
        """
        Retrieve the current chat settings for the broadcaster.

        ???+ note
            The scopes only required if you want to include non_moderator extra keys.

        | Scopes                           | Description                                |
        | -------------------------------- | -------------------------------------------|
        | `moderator:read:chat_settings`   | View a broadcaster’s chat room settings.   |
        | `moderator:manage:chat_settings` | Manage a broadcaster’s chat room settings. |

        Returns
        -------
        chat.Settings
            A dictionary containing chat settings for the broadcaster.
        """
        data: Data[List[chat.Settings]] = await self._state.http.get_chat_settings(self._user_id,
                                                                                   self._auth_user_id)
        return data['data'][0]

    async def update_settings(self,
                              emote_mode: Optional[bool] = None,
                              follower_mode: Optional[bool] = None,
                              follower_mode_duration: Optional[int] = None,
                              non_moderator_chat_delay: Optional[bool] = None,
                              non_moderator_chat_delay_duration: Optional[int] = None,
                              slow_mode: Optional[bool] = None,
                              slow_mode_wait_time: Optional[int] = None,
                              subscriber_mode: Optional[bool] = None,
                              unique_chat_mode: Optional[bool] = None) -> chat.Settings:
        """
        Update the chat settings for the broadcaster.

        | Scopes                           | Description                                |
        | -------------------------------- | -------------------------------------------|
        | `moderator:manage:chat_settings` | Manage a broadcaster’s chat room settings. |

        Parameters
        ----------
        emote_mode: Optional[bool]
            Whether to enable emote-only mode.
        follower_mode: Optional[bool]
            Whether to enable follower-only mode.
        follower_mode_duration: Optional[int]
            Duration of the follower-only mode in minutes.
        non_moderator_chat_delay: Optional[bool]
            Whether to enable chat delay for non-moderators.
        non_moderator_chat_delay_duration: Optional[int]
            Duration of the chat delay for non-moderators in seconds.
        slow_mode: Optional[bool]
            Whether to enable slow mode.
        slow_mode_wait_time: Optional[int]
            The duration of the slow mode in seconds.
        subscriber_mode: Optional[bool]
            Whether to enable subscriber-only mode.
        unique_chat_mode: Optional[bool]
            Whether to enable unique chat mode.

        Returns
        -------
        chat.Settings
            A dictionary containing the updated chat settings.
        """
        kwargs: Dict[str, Any] = {
            'emote_mode': emote_mode,
            'follower_mode': follower_mode,
            'follower_mode_duration': follower_mode_duration if follower_mode else None,
            'non_moderator_chat_delay': non_moderator_chat_delay,
            'non_moderator_chat_delay_duration': (non_moderator_chat_delay_duration
                                                  if non_moderator_chat_delay else None),
            'slow_mode': slow_mode,
            'slow_mode_wait_time': slow_mode_wait_time if slow_mode else None,
            'subscriber_mode': subscriber_mode,
            'unique_chat_mode': unique_chat_mode
        }
        data: Data[List[chat.Settings]] = await self._state.http.update_chat_settings(self._user_id,
                                                                                      self._auth_user_id,
                                                                                      **kwargs)
        return data['data'][0]

    async def get_total_chatters(self) -> int:
        """
        Get the total number of chatters in the broadcaster's chat.

        | Scopes                    | Description                                     |
        | ------------------------- | ------------------------------------------------|
        | `moderator:read:chatters` | View the chatters in a broadcaster’s chat room. |

        Returns
        -------
        int
            The total number of chatters.
        """
        data: TData[List[users.SpecificUser]] = await self._state.http.get_chatters(self._user_id,
                                                                                    self._auth_user_id,
                                                                                    first=1)
        return data['total']

    async def fetch_chatters(self, *, first: int = 100) -> AsyncGenerator[List[users.SpecificUser], None]:
        """
        Fetch a list of chatters in the broadcaster's chat.

        | Scopes                    | Description                                     |
        | ------------------------- | ------------------------------------------------|
        | `moderator:read:chatters` | View the chatters in a broadcaster’s chat room. |

        Parameters
        ----------
        first: int
            The number of chatters to fetch per request.

        Yields
        ------
        AsyncGenerator[List[users.SpecificUser], None]
            A list of dictionaries representing chatters.
        """
        kwargs: Dict[str, Any] = {
            'broadcaster_id': self._user_id,
            'moderator_id': self._auth_user_id,
            'first': first,
            'after': None
        }
        while True:
            data: TData[List[users.SpecificUser]] = await self._state.http.get_chatters(**kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def get_emotes(self) -> Tuple[List[chat.Emote], str]:
        """
        Retrieve the emotes and template for the broadcaster's channel.

        | Scopes                    | Description                                     |
        | ------------------------- | ------------------------------------------------|
        | `moderator:read:chatters` | View the chatters in a broadcaster’s chat room. |

        Returns
        -------
        Tuple[List[chat.Emote], str]
            A tuple where the first element is a list of dictionaries representing emotes,
            and the second element is a string representing the template.
        """
        data: Edata[List[chat.Emote]] = await self._state.http.get_channel_emotes(self._auth_user_id, self._user_id)
        return data['data'], data['template']

    async def get_cheermotes(self) -> List[bits.Cheermote]:
        """
        Retrieve the cheermotes for the broadcaster's channel.

        Returns
        -------
        List[bits.Cheermote]
            A list of dictionaries representing cheermotes.
        """
        data: Data[List[bits.Cheermote]] = await self._state.http.get_cheermotes(self._auth_user_id, self._user_id)
        return data['data']

    async def get_badges(self) -> List[chat.Badge]:
        """
        Retrieve the chat badges for the broadcaster's channel.

        Returns
        -------
        List[chat.Badge]
            A list of dictionaries representing chat badges.
        """
        data: Data[List[chat.Badge]] = await self._state.http.get_channel_chat_badges(self._auth_user_id, self._user_id)
        return data['data']

    async def send_shoutout(self, user: User) -> None:
        """
        Send a shoutout to another broadcaster.

        | Scopes                       | Description                       |
        | ---------------------------- | ----------------------------------|
        | `moderator:manage:shoutouts` | Manage a broadcaster’s shoutouts. |

        Parameters
        ----------
        user: User
            The user to send a shoutout to.
        """
        await self._state.http.send_a_shoutout(self._user_id, self._auth_user_id, to_broadcaster_id=user.id)

    async def send_announcement(self,
                                message: str,
                                *,
                                color: Literal['blue', 'green', 'orange', 'purple', 'primary'] = 'primary'):
        """
        Send an announcement message to the chat.

        | Scopes                           | Description                                                       |
        | -------------------------------- | ------------------------------------------------------------------|
        | `moderator:manage:announcements` | Send announcements in channels where you have the moderator role. |

        Parameters
        ----------
        message: str
            The announcement message to send.
        color: Literal['blue', 'green', 'orange', 'purple', 'primary']
            The color of the announcement message.
        """
        await self._state.http.send_chat_announcement(self._user_id, self._auth_user_id, message=message, color=color)

    async def warn_user(self, user: User, reason: str) -> None:
        """
        Send a warning to a user in the chat.

        | Scopes                      | Description                                               |
        | --------------------------- | ----------------------------------------------------------|
        | `moderator:manage:warnings` | Warn users in channels where you have the moderator role. |

        Parameters
        ----------
        user: User
            The user to warn.
        reason: str
            The reason for the warning.
        """
        await self._state.http.warn_chat_user(self._user_id, self._auth_user_id, user_id=user.id, reason=reason)

    async def send_message(self, text: str, reply_message_id: Optional[str] = None) -> chat.SendMessageStatus:
        """
        Send a message to the chat.

        ???+ warning
            It's strongly discouraged to send messages to other channels without the owner's permission.
            Doing so could result in a permanent ban from twitch. Please be cautious.

        | Scopes           | Description                                                     |
        | ---------------- | ----------------------------------------------------------------|
        | `user:read:chat` | Join a specified chat channel as your user and appear as a bot. |

        Parameters
        ----------
        text: str
            The message text to send.
        reply_message_id: Optional[str]
            The ID of the message to reply to.

        Returns
        -------
        chat.SendMessageStatus
            A dictionary containing the status of the user sent message.
        """
        data: Data[List[chat.SendMessageStatus]] = await self._state.http.send_chat_message(
            self._user_id,
            self._auth_user_id,
            text=text,
            reply_parent_message_id=reply_message_id)
        return data['data'][0]

    async def delete_message(self, msg_id: str) -> None:
        """
        Delete a message from the chat.

        | Scopes                           | Description                                                         |
        | -------------------------------- | --------------------------------------------------------------------|
        | `moderator:manage:chat_messages` | Delete chat messages in channels where you have the moderator role. |

        Parameters
        ----------
        msg_id: str
            The ID of the message to delete.
        """
        await self._state.http.delete_chat_messages(self._user_id, self._auth_user_id, message_id=msg_id)

    async def clear_chat(self) -> None:
        """
        Clear all messages from the chat.

        | Scopes                           | Description                                                         |
        | -------------------------------- | --------------------------------------------------------------------|
        | `moderator:manage:chat_messages` | Delete chat messages in channels where you have the moderator role. |
        """
        await self._state.http.delete_chat_messages(self._user_id, self._auth_user_id)

    async def get_shieldmode(self) -> moderation.ShieldModeStatus:
        """
        Retrieve the current shield mode status for the broadcaster's channel.

        | Scopes                         | Description                                |
        | ------------------------------ | -------------------------------------------|
        | `moderator:read:shield_mode`   | View a broadcaster’s Shield Mode status.   |
        | `moderator:manage:shield_mode` | Manage a broadcaster’s Shield Mode status. |


        Returns
        -------
        moderation.ShieldModeStatus
            A dictionary representing the shield mode status.
        """
        data: Data[List[moderation.ShieldModeStatus]] = await self._state.http.get_shield_mode_status(
            self._user_id,
            self._auth_user_id)
        return data['data'][0]

    async def update_shieldmode(self, activate: bool) -> moderation.ShieldModeStatus:
        """
        Update the shield mode status for the broadcaster's channel.

        | Scopes                         | Description                                |
        | ------------------------------ | -------------------------------------------|
        | `moderator:manage:shield_mode` | Manage a broadcaster’s Shield Mode status. |

        Parameters
        ----------
        activate: bool
            Whether to activate or deactivate shield mode.

        Returns
        -------
        moderation.ShieldModeStatus
            A dictionary representing the updated shield mode status.
        """
        data: Data[List[moderation.ShieldModeStatus]] = await self._state.http.update_shield_mode_status(
            self._user_id,
            self._auth_user_id,
            is_active=activate)
        return data['data'][0]

    async def get_automod_settings(self) -> moderation.AutoModSettings:
        """
        Retrieve the current AutoMod settings for the broadcaster's channel.

        | Scopes                              | Description                              |
        | ----------------------------------- | -----------------------------------------|
        | `moderator:read:automod_settings`   | View a broadcaster’s AutoMod settings.   |
        | `moderator:manage:automod_settings` | Manage a broadcaster’s AutoMod settings. |

        Returns
        -------
        moderation.AutoModSettings
            A dictionary representing the AutoMod settings.
        """
        data: Data[List[moderation.AutoModSettings]] = await self._state.http.get_automod_settings(self._user_id,
                                                                                                   self._auth_user_id)
        return data['data'][0]

    async def update_automod_settings(self,
                                      aggression: Optional[int] = None,
                                      bullying: Optional[int] = None,
                                      disability: Optional[int] = None,
                                      misogyny: Optional[int] = None,
                                      race_ethnicity_or_religion: Optional[int] = None,
                                      sex_based_terms: Optional[int] = None,
                                      sexuality_sex_or_gender: Optional[int] = None,
                                      swearing: Optional[int] = None) -> moderation.AutoModSettings:
        """
        Update the AutoMod settings for the broadcaster's channel.

        | Scopes                              | Description                              |
        | ----------------------------------- | -----------------------------------------|
        | `moderator:manage:automod_settings` | Manage a broadcaster’s AutoMod settings. |

        Parameters
        ----------
        aggression: Optional[int]
            The level of aggression moderation.
        bullying: Optional[int]
            The level of bullying moderation.
        disability: Optional[int]
            The level of disability-related moderation.
        misogyny: Optional[int]
            The level of misogyny moderation.
        race_ethnicity_or_religion: Optional[int]
            The level of race, ethnicity, or religion-related moderation.
        sex_based_terms: Optional[int]
            The level of sex-based terms moderation.
        sexuality_sex_or_gender: Optional[int]
            The level of sexuality, sex, or gender-related moderation.
        swearing: Optional[int]
            The level of swearing moderation.

        Returns
        -------
        moderation.AutoModSettings
            A dictionary representing the updated AutoMod settings.
        """
        kwargs: Dict[str, Any] = {
            'aggression': aggression,
            'bullying': bullying,
            'disability': disability,
            'misogyny': misogyny,
            'race_ethnicity_or_religion': race_ethnicity_or_religion,
            'sex_based_terms': sex_based_terms,
            'sexuality_sex_or_gender': sexuality_sex_or_gender,
            'swearing': swearing
        }
        data: Data[List[moderation.AutoModSettings]] = await self._state.http.update_automod_settings(
            self._user_id,
            self._auth_user_id,
            **kwargs)
        return data['data'][0]

    async def update_automod_settings_level(self, overall_level: int) -> moderation.AutoModSettings:
        """
        Update the overall AutoMod settings level for the broadcaster's channel.

        | Scopes                              | Description                              |
        | ----------------------------------- | -----------------------------------------|
        | `moderator:manage:automod_settings` | Manage a broadcaster’s AutoMod settings. |

        Parameters
        ----------
        overall_level: int
            The new overall AutoMod settings level.

        Returns
        -------
        moderation.AutoModSettings
            A dictionary representing the updated AutoMod settings.
        """
        data: Data[List[moderation.AutoModSettings]] = await self._state.http.update_automod_settings(
            self._user_id,
            self._auth_user_id,
            overall_level=overall_level)
        return data['data'][0]

    async def automod_held_message(self, message_id: str, action: Literal['allow', 'deny']) -> None:
        """
        Manage a held AutoMod message by either allowing or denying it.

        | Scopes                     | Description                                 |
        | -------------------------- | --------------------------------------------|
        | `moderator:manage:automod` | Manage messages held for review by AutoMod. |

        Parameters
        ----------
        message_id: str
            The ID of the message to manage.

        action: Literal['allow', 'deny']
            The action to perform on the held message; either 'allow' or 'deny'.
        """
        await self._state.http.manage_held_automod_messages(self._auth_user_id, message_id, action)

    async def fetch_blocked_terms(self, first: int = 100) -> AsyncGenerator[List[moderation.BlockedTerm], None]:
        """
        Retrieve blocked terms for the broadcaster's channel in pages.

        | Scopes                           | Description                                   |
        | -------------------------------- | ----------------------------------------------|
        | `moderator:read:blocked_terms`   | View a broadcaster’s list of blocked terms.   |
        | `moderator:manage:blocked_terms` | Manage a broadcaster’s list of blocked terms. |

        Parameters
        ----------
        first: int
            The maximum number of blocked terms to retrieve per page.

        Yields
        ------
        AsyncGenerator[List[moderation.BlockedTerm], None]
            A list of dictionaries, each representing a blocked term.
        """
        kwargs: Dict[str, Any] = {
            'broadcaster_id': self._user_id,
            'moderator_id': self._auth_user_id,
            'first': first,
            'after': None
        }
        while True:
            data: PData[List[moderation.BlockedTerm]] = await self._state.http.get_blocked_terms(**kwargs)
            yield data['data']
            kwargs['after'] = data['pagination'].get('cursor')
            if not kwargs['after']:
                break

    async def add_blocked_term(self, text: str) -> moderation.BlockedTerm:
        """
        Add a term to the list of blocked terms for the broadcaster's channel.

        | Scopes                           | Description                                   |
        | -------------------------------- | ----------------------------------------------|
        | `moderator:manage:blocked_terms` | Manage a broadcaster’s list of blocked terms. |

        Parameters
        ----------
        text: str
            The term to be blocked.

        Returns
        -------
        moderation.BlockedTerm
            A dictionary representing the blocked term that was added.
        """
        data: Data[List[moderation.BlockedTerm]] = await self._state.http.add_blocked_term(self._user_id,
                                                                                           self._auth_user_id,
                                                                                           text=text)
        return data['data'][0]

    async def remove_blocked_term(self, term_id: str) -> None:
        """
        Remove a term from the list of blocked terms for the broadcaster's channel.

        | Scopes                           | Description                                   |
        | -------------------------------- | ----------------------------------------------|
        | `moderator:manage:blocked_terms` | Manage a broadcaster’s list of blocked terms. |

        Parameters
        ----------
        term_id: str
            The ID of the blocked term to be removed.
        """
        await self._state.http.remove_blocked_term(self._user_id, self._auth_user_id, term_id=term_id)


class BroadcasterChat(Chat):
    """
    Represents a Broadcaster channel chat.
    """
    __slots__ = ()

    def __init__(self, user_id: str, *, state: ConnectionState) -> None:
        super().__init__(user_id, user_id, state=state)

    async def automod_check_messages(self, messages: List[str]) -> List[moderation.AutoModMessageStatus]:
        """
        Check the AutoMod status of multiple messages.

        | Scopes            | Description                  |
        | ----------------- | -----------------------------|
        | `moderation:read` | View a channel’s moderation. |

        Parameters
        ----------
        messages: List[str]
            A list of messages to check.

        Returns
        -------
        List[moderation.AutoModMessageStatus]
            A list of dictionaries, each representing the AutoMod status of a message.
        """
        data: Data[List[moderation.AutoModMessageStatus]] = await self._state.http.check_automod_status(
            self._state.user.id,
            messages)
        return data['data']



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TypedDict, TYPE_CHECKING

if TYPE_CHECKING:
    from typing import List


# Game Search
class Game(TypedDict):
    """
    Represents a game with its details.

    Attributes
    ----------
    id: str
        The unique identifier for the game.
    name: str
        The name of the game.
    box_art_url: str
        The URL of the game's box art.
    igdb_id: str
        The IGDB (Internet Game Database) identifier for the game.
    """
    id: str
    name: str
    box_art_url: str
    igdb_id: str


# Category Search
class CategorySearch(TypedDict):
    """
    Represents a category with its details.

    Attributes
    ----------
    box_art_url: str
        The URL of the category's box art.
    name: str
        The name of the category.
    id: str
        The unique identifier for the category.
    """
    box_art_url: str
    name: str
    id: str


# Channel Search
class ChannelSearch(TypedDict):
    """
    Represents a channel with its details.

    Attributes
    ----------
    broadcaster_language: str
        The language spoken by the broadcaster.
    broadcaster_login: str
        The login name of the broadcaster.
    display_name: str
        The display name of the channel.
    game_id: str
        The unique identifier of the game being played.
    game_name: str
        The name of the game being played.
    id: str
        The unique identifier for the channel.
    is_live: bool
        Indicates whether the channel is currently live.
    tags: List[str]
        A list of tags associated with the channel.
    thumbnail_url: str
        The URL of the channel's thumbnail.
    title: str
        The title of the channel's current broadcast.
    started_at: str
        The start time of the channel's current broadcast in ISO 8601 format.
    """
    broadcaster_language: str
    broadcaster_login: str
    display_name: str
    game_id: str
    game_name: str
    id: str
    is_live: bool
    tags: List[str]
    thumbnail_url: str
    title: str
    started_at: str



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from .users import SpecificBroadcaster, SpecificUser, SpecificModerator, Moderator
from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from typing import List, Optional, Literal


# AutoMod
class Emote(TypedDict):
    """
    Represents an emote.

    Attributes
    ----------
    id: str
        The ID of the emote.
    emote_set_id: str
        The ID of the emote set to which the emote belongs.
    """
    id: str
    emote_set_id: str


class Cheermote(TypedDict):
    """
    Represents a cheermote.

    Attributes
    ----------
    prefix: str
        The prefix of the cheermote.
    bits: int
        The number of bits required to use the cheermote.
    tier: int
        The tier level of the cheermote.
    """
    prefix: str
    bits: int
    tier: int


class Fragment(TypedDict):
    """
    Represents a fragment of a message that can be an emote or cheermote.

    Attributes
    ----------
    text: str
        The text content of the fragment.
    emote: Optional[Emote]
        The emote in the fragment, if any.
    cheermote: Optional[Cheermote]
        The cheermote in the fragment, if any.
    """
    text: str
    emote: Optional[Emote]
    cheermote: Optional[Cheermote]


class Message(TypedDict):
    """
    Represents a message with its content and fragments.

    Attributes
    ----------
    text: str
        The text content of the message.
    fragments: List[Fragment]
        A list of fragments in the message.
    """
    text: str
    fragments: List[Fragment]


class AutomodMessageHoldEvent(SpecificBroadcaster, SpecificUser):
    """
    Represents an event where a message is held by AutoMod.

    Attributes
    ----------
    message_id: str
        The ID of the held message.
    message: Message
        The details of the held message.
    category: str
        The category of the AutoMod action.
    level: int
        The moderation level of the AutoMod action.
    held_at: str
        The timestamp when the message was held, in ISO 8601 format.
    """
    message_id: str
    message: Message
    category: str
    level: int
    held_at: str


class AutomodMessageUpdateEvent(SpecificBroadcaster, SpecificModerator, SpecificUser):
    """
    Represents an event where a held message is updated by AutoMod.

    Attributes
    ----------
    message_id: str
        The ID of the updated message.
    message: Message
        The updated message details.
    category: str
        The category of the AutoMod action.
    level: int
        The moderation level of the AutoMod action.
    status: Literal['Approved', 'Denied', 'Expired']
        The current status of the message.
    held_at: str
        The timestamp when the message was held, in ISO 8601 format.
    """
    message_id: str
    message: Message
    category: str
    level: int
    status: Literal['Approved', 'Denied', 'Expired']
    held_at: str


class AutomodSettingsUpdateEvent(SpecificBroadcaster, SpecificModerator):
    """
    Represents an event where AutoMod settings are updated.

    Attributes
    ----------
    bullying: int
        The level of moderation for bullying.
    overall_level: Optional[int]
        The overall moderation level, if applicable.
    disability: int
        The level of moderation for disability-related terms.
    race_ethnicity_or_religion: int
        The level of moderation for race, ethnicity, or religion.
    misogyny: int
        The level of moderation for misogyny.
    sexuality_sex_or_gender: int
        The level of moderation for sexuality, sex, or gender.
    aggression: int
        The level of moderation for aggression.
    sex_based_terms: int
        The level of moderation for sex-based terms.
    swearing: int
        The level of moderation for swearing.
    """
    bullying: int
    overall_level: Optional[int]
    disability: int
    race_ethnicity_or_religion: int
    misogyny: int
    sexuality_sex_or_gender: int
    aggression: int
    sex_based_terms: int
    swearing: int


class AutomodTermsUpdateEvent(SpecificBroadcaster, SpecificModerator):
    """
    Represents an event where AutoMod terms are updated.

    Attributes
    ----------
    action: Literal['add_permitted', 'remove_permitted', 'add_blocked', 'remove_blocked']
        The action taken on the terms.
    from_automod: bool
        Whether the terms update is from AutoMod.
    terms: List[str]
        A list of terms involved in the update.
    """
    action: Literal['add_permitted', 'remove_permitted', 'add_blocked', 'remove_blocked']
    from_automod: bool
    terms: List[str]


class BanEvent(SpecificBroadcaster, SpecificModerator, SpecificUser):
    """
    Represents an event where a user is banned.

    Attributes
    ----------
    reason: str
        The reason for the ban.
    banned_at: str
        The timestamp when the ban was applied, in ISO 8601 format.
    ends_at: Optional[str]
        The timestamp when the ban will end, if applicable.
    is_permanent: bool
        Whether the ban is permanent.
    """
    reason: str
    banned_at: str
    ends_at: Optional[str]
    is_permanent: bool


class UnbanEvent(SpecificBroadcaster, SpecificModerator, SpecificUser):
    """
    Represents an event where a user is unbanned.

    Attributes
    ----------
    (No additional attributes; inherits from SpecificBroadcaster, SpecificModerator, SpecificUser)
    """
    pass


class UnbanRequestCreateEvent(SpecificBroadcaster, SpecificUser):
    """
    Represents an event where an unban request is created.

    Attributes
    ----------
    id: str
        The ID of the unban request.
    text: str
        The text of the unban request.
    created_at: str
        The timestamp when the unban request was created, in ISO 8601 format.
    """
    id: str
    text: str
    created_at: str


class UnbanRequestResolveEvent(SpecificBroadcaster, Moderator, SpecificUser):
    """
    Represents an event where an unban request is resolved.

    Attributes
    ----------
    id: str
        The ID of the unban request.
    resolution_text: Optional[str]
        The resolution text or reason for the decision.
    status: str
        The status of the unban request.
    """
    id: str
    resolution_text: Optional[str]
    status: str


class FollowersMetadata(TypedDict):
    """
    Represents metadata related to followers.

    Attributes
    ----------
    follow_duration_minutes: int
        The duration for which the user has been following, in minutes.
    """
    follow_duration_minutes: int


class SlowMetadata(TypedDict):
    """
    Represents metadata related to slow mode.

    Attributes
    ----------
    wait_time_seconds: int
        The wait time for slow mode, in seconds.
    """
    wait_time_seconds: int


class BanMetadata(SpecificUser):
    """
    Represents metadata related to a ban.

    Attributes
    ----------
    reason: Optional[str]
        The reason for the ban, if provided.
    """
    reason: Optional[str]


class TimeoutMetadata(SpecificUser):
    """
    Represents metadata related to a timeout.

    Attributes
    ----------
    reason: Optional[str]
        The reason for the timeout, if provided.
    expires_at: Optional[str]
        The timestamp when the timeout will end, in ISO 8601 format.
    """
    reason: Optional[str]
    expires_at: Optional[str]


class RaidMetadata(SpecificUser):
    """
    Represents metadata related to a raid.

    Attributes
    ----------
    viewer_count: int
        The number of viewers involved in the raid.
    """
    viewer_count: int


class DeleteMetadata(SpecificUser):
    """
    Represents metadata related to a message deletion.

    Attributes
    ----------
    message_id: str
        The ID of the deleted message.
    message_body: str
        The body of the deleted message.
    """
    message_id: str
    message_body: str


class AutomodTermsMetadata(TypedDict):
    """
    Represents metadata related to AutoMod terms updates.

    Attributes
    ----------
    action: Literal['add', 'remove']
        The action taken on the terms.
    list: Literal['blocked', 'permitted']
        The list to which the terms were added or removed.
    terms: List[str]
        A list of terms involved in the update.
    from_automod: bool
        Whether the terms update is from AutoMod.
    """
    action: Literal['add', 'remove']
    list: Literal['blocked', 'permitted']
    terms: List[str]
    from_automod: bool


class UnbanRequestMetadata(SpecificUser):
    """
    Represents metadata related to an unban request.

    Attributes
    ----------
    is_approved: bool
        Whether the unban request was approved.
    moderator_message: Optional[str]
        The message from the moderator regarding the unban request.
    """
    is_approved: bool
    moderator_message: Optional[str]


class WarnMetadata(SpecificUser):
    """
    Represents metadata related to a warning.

    Attributes
    ----------
    reason: Optional[str]
        The reason for the warning, if provided.
    chat_rules_cited: Optional[List[str]]
        A list of chat rules cited in the warning, if any.
    """
    reason: Optional[str]
    chat_rules_cited: Optional[List[str]]


class ModerateEvent(SpecificBroadcaster, SpecificModerator):
    """
    Represents an event where moderation actions are taken.

    Attributes
    ----------
    action: str
        The type of moderation action.
    followers: Optional[FollowersMetadata]
        Metadata related to followers, if applicable.
    slow: Optional[SlowMetadata]
        Metadata related to slow mode, if applicable.
    vip: Optional[SpecificUser]
        The user who was made a VIP, if applicable.
    unvip: Optional[SpecificUser]
        The user who was removed from VIP status, if applicable.
    mod: Optional[SpecificUser]
        The user who was given moderator status, if applicable.
    unmod: Optional[SpecificUser]
        The user who was removed from moderator status, if applicable.
    ban: Optional[BanMetadata]
        Metadata related to a ban, if applicable.
    unban: Optional[SpecificUser]
        The user who was unbanned, if applicable.
    timeout: Optional[TimeoutMetadata]
        Metadata related to a timeout, if applicable.
    untimeout: Optional[SpecificUser]
        The user who was removed from timeout status, if applicable.
    raid: Optional[RaidMetadata]
        Metadata related to a raid, if applicable.
    unraid: Optional[SpecificUser]
        The user who was removed from raid status, if applicable.
    delete: Optional[DeleteMetadata]
        Metadata related to a message deletion, if applicable.
    automod_terms: Optional[AutomodTermsMetadata]
        Metadata related to AutoMod terms updates, if applicable.
    unban_request: Optional[UnbanRequestMetadata]
        Metadata related to an unban request, if applicable.
    warn: Optional[WarnMetadata]
        Metadata related to a warning, if applicable.
    shared_chat_ban: Optional[BanMetadata]
        Information about the shared_chat_ban event, if applicable.
    shared_chat_unban: Optional[SpecificUser]
        Information about the shared_chat_unban event, if applicable.
    shared_chat_timeout: Optional[TimeoutMetadata]
        Information about the shared_chat_timeout event, if applicable.
    shared_chat_untimeout: Optional[SpecificUser]
        Information about the shared_chat_untimeout event, if applicable.
    shared_chat_delete: Optional[DeleteMetadata]
        Information about the shared_chat_delete event, if applicable.
    """
    action: str
    followers: Optional[FollowersMetadata]
    slow: Optional[SlowMetadata]
    vip: Optional[SpecificUser]
    unvip: Optional[SpecificUser]
    mod: Optional[SpecificUser]
    unmod: Optional[SpecificUser]
    ban: Optional[BanMetadata]
    unban: Optional[SpecificUser]
    timeout: Optional[TimeoutMetadata]
    untimeout: Optional[SpecificUser]
    raid: Optional[RaidMetadata]
    unraid: Optional[SpecificUser]
    delete: Optional[DeleteMetadata]
    automod_terms: Optional[AutomodTermsMetadata]
    unban_request: Optional[UnbanRequestMetadata]
    warn: Optional[WarnMetadata]
    shared_chat_ban: Optional[BanMetadata]
    shared_chat_unban: Optional[SpecificUser]
    shared_chat_timeout: Optional[TimeoutMetadata]
    shared_chat_untimeout: Optional[SpecificUser]
    shared_chat_delete: Optional[DeleteMetadata]


# Moderator Add/Remove
class ModeratorAddEvent(SpecificBroadcaster, SpecificUser):
    """
    Represents an event where a user is added as a moderator.

    Attributes
    ----------
    (No additional attributes; inherits from SpecificBroadcaster, SpecificUser)
    """
    pass


class ModeratorRemoveEvent(SpecificBroadcaster, SpecificUser):
    """
    Represents an event where a user is removed as a moderator.

    Attributes
    ----------
    (No additional attributes; inherits from SpecificBroadcaster, SpecificUser)
    """
    pass


# SuspiciousUser
class ChatFragment(TypedDict):
    """
    Represents a fragment of a chat message.

    Attributes
    ----------
    type: Literal['text', 'cheermote', 'emote']
        The type of fragment.
    text: str
        The text content of the fragment.
    cheermote: Optional[Cheermote]
        The cheermote in the fragment, if any.
    emote: Optional[Emote]
        The emote in the fragment, if any.
    """
    type: Literal['text', 'cheermote', 'emote']
    text: str
    cheermote: Optional[Cheermote]
    emote: Optional[Emote]


class ChatMessage(TypedDict):
    """
    Represents a chat message with its fragments.

    Attributes
    ----------
    message_id: str
        The ID of the chat message.
    text: str
        The text content of the chat message.
    fragments: List[ChatFragment]
        A list of fragments in the chat message.
    """
    message_id: str
    text: str
    fragments: List[ChatFragment]


class SuspiciousUserUpdateEvent(SpecificBroadcaster, SpecificModerator, SpecificUser):
    """
    Represents an update to a suspicious user's status.

    Attributes
    ----------
    low_trust_status: Literal['none', 'active_monitoring', 'restricted']
        The current low trust status of the user.
    """
    low_trust_status: Literal['none', 'active_monitoring', 'restricted']


class SuspiciousUserMessageEvent(SpecificBroadcaster, SpecificUser):
    """
    Represents an event where a suspicious user sends a message.

    Attributes
    ----------
    low_trust_status: Literal['none', 'active_monitoring', 'restricted']
        The current low trust status of the user.
    shared_ban_channel_ids: List[str]
        A list of channel IDs where the user has shared bans.
    types: List[Literal['manual', 'ban_evader_detector', 'shared_channel_ban']]
        A list of types associated with the suspicious behavior.
    ban_evasion_evaluation: Literal['unknown', 'possible', 'likely']
        The evaluation of ban evasion.
    message: ChatMessage
        The message sent by the suspicious user.
    """
    low_trust_status: Literal['none', 'active_monitoring', 'restricted']
    shared_ban_channel_ids: List[str]
    types: List[Literal['manual', 'ban_evader_detector', 'shared_channel_ban']]
    ban_evasion_evaluation: Literal['unknown', 'possible', 'likely']
    message: ChatMessage


# Warning
class WarningAcknowledgeEvent(SpecificBroadcaster, SpecificUser):
    """
    Represents an event where a warning is acknowledged.

    Attributes
    ----------
    (No additional attributes; inherits from SpecificBroadcaster, SpecificUser)
    """
    pass


class WarningSendEvent(SpecificBroadcaster, SpecificModerator, SpecificUser):
    """
    Represents an event where a warning is sent.

    Attributes
    ----------
    reason: Optional[str]
        The reason for the warning, if provided.
    chat_rules_cited: Optional[List[str]]
        A list of chat rules cited in the warning, if any.
    """
    reason: Optional[str]
    chat_rules_cited: Optional[List[str]]


# ShieldMode
class ShieldModeBeginEvent(SpecificBroadcaster, SpecificModerator):
    """
    Represents an event where Shield Mode begins.

    Attributes
    ----------
    started_at: str
        The timestamp when Shield Mode began, in ISO 8601 format.
    """
    started_at: str


class ShieldModeEndEvent(SpecificBroadcaster, SpecificModerator):
    """
    Represents an event where Shield Mode ends.

    Attributes
    ----------
    ended_at: str
        The timestamp when Shield Mode ended, in ISO 8601 format.
    """
    ended_at: str


"""
twitch.py

Real-time Twitch Event Handling and Helix API Integration in Python

:copyright: (c) 2024-present Snifo
:license: MIT, see LICENSE for more details.
"""

__title__ = 'twitch.py'
__version__ = '3.2.3'
__license__ = 'MIT License'
__author__ = 'Snifo'
__email__ = 'Snifo@mail.com'
__github__ = 'https://github.com/mrsnifo/twitch.py'

from .user import *
from .chat import *
from .utils import *
from .client import *
from .errors import *
from .stream import *
from .channel import *



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING
import gzip
import io

if TYPE_CHECKING:
    from typing import Optional, Dict

import logging
_logger = logging.getLogger(__name__)


class Attachment:
    """
    Manages file attachments in memory, including loading, compressing,
    decompressing, and retrieving files.
    """

    __slots__ = ('attachments', 'path_keys')

    def __init__(self) -> None:
        self.attachments: Dict[str, bytes] = {}
        self.path_keys: Dict[str, str] = {}

    def clear(self) -> None:
        """Clears all attachments and path keys from memory."""
        self.attachments.clear()
        self.path_keys.clear()
        _logger.debug('All attachments have been cleared from memory.')

    @staticmethod
    def compress_data(data: bytes) -> bytes:
        """Compresses data using gzip."""
        buffer = io.BytesIO()
        with gzip.GzipFile(fileobj=buffer, mode='wb') as gzip_file:
            gzip_file.write(data)
        return buffer.getvalue()

    @staticmethod
    def decompress_data(compressed_data: bytes) -> bytes:
        """Decompresses gzip-compressed data."""
        buffer = io.BytesIO(compressed_data)
        with gzip.GzipFile(fileobj=buffer, mode='rb') as gzip_file:
            return gzip_file.read()

    @staticmethod
    def convert_path_to_key(path: str) -> str:
        """Converts a file path to a hash-based key with the file extension."""
        ext = f'.{path.rsplit(".", 1)[-1]}' if '.' in path else ''
        key = f"{abs(hash(path))}{ext}"
        _logger.debug('Converted path `%s` to key `%s`.', path, key)
        return key

    def load_file_into_memory(self, name: str, path: str) -> None:
        """Loads a file from the filesystem into memory and compresses it."""
        with open(path, 'rb') as file:
            file_data = file.read()
            compressed_data = self.compress_data(file_data)
            path_key = self.convert_path_to_key(path)
            self.attachments[path_key] = compressed_data
            self.path_keys[name] = path_key
        _logger.debug('File `%s` loaded into memory with name `%s`.', path, name)

    def get_attachment(self, path_key: str) -> Optional[bytes]:
        """Retrieves and decompresses a file from memory."""
        compressed_file_content = self.attachments.get(path_key)
        if compressed_file_content:
            return self.decompress_data(compressed_file_content)
        return None

    def get_path_key(self, name: str) -> Optional[str]:
        """Retrieves the path key for a given name."""
        return self.path_keys.get(name, name)

    def remove_attachment(self, name: str) -> None:
        """Removes a file from memory based on its name."""
        try:
            if name in self.path_keys:
                path_key = self.path_keys.pop(name)
                self.attachments.pop(path_key, None)
                _logger.debug('Removed attachment with name `%s`.', name)
        except Exception as exc:
            _logger.exception('Failed to remove attachment', exc)



"""
twitch.types.eventsub

Typing for Twitch.

:copyright: (c) 2024-present Snifo
:license: MIT, see LICENSE for more details.
"""

from __future__ import annotations

from . import activity, bits, channels, chat, interaction, moderation, streams, users
from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from typing import TypeVar, Dict, Any
    T = TypeVar('T')


class Subscription(TypedDict):
    id: str
    type: str
    version: str
    status: str
    cost: int
    condition: Dict[str, Any]
    created_at: str


class Data(TypedDict):
    subscription: Subscription
    event: T



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from .users import SpecificBroadcaster, Broadcaster, SpecificUser
from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from typing import Literal


class DonationAmount(TypedDict):
    """
    Represents the amount of a donation.

    Attributes
    ----------
    value: int
        The value of the donation.
    decimal_places: int
        The number of decimal places in the donation value.
    currency: str
        The currency code (e.g., USD) of the donation.
    """
    value: int
    decimal_places: int
    currency: str


class CharityDonationEvent(SpecificBroadcaster, SpecificUser):
    """
    Represents an event where a donation is made to a charity during a Twitch stream.

    Attributes
    ----------
    id: str
        The unique identifier for the charity donation event.
    campaign_id: str
        The unique identifier for the charity campaign.
    charity_name: str
        The name of the charity.
    charity_description: str
        A description of the charity.
    charity_logo: str
        The URL of the charity's logo.
    charity_website: str
        The URL of the charity's website.
    amount: DonationAmount
        The amount donated to the charity.
    """
    id: str
    campaign_id: str
    charity_name: str
    charity_description: str
    charity_logo: str
    charity_website: str
    amount: DonationAmount


class CharityCampaignStartEvent(Broadcaster):
    """
    Represents the start of a charity campaign during a Twitch stream.

    Attributes
    ----------
    id: str
        The unique identifier for the charity campaign.
    charity_name: str
        The name of the charity.
    charity_description: str
        A description of the charity.
    charity_logo: str
        The URL of the charity's logo.
    charity_website: str
        The URL of the charity's website.
    current_amount: DonationAmount
        The current amount raised during the campaign.
    target_amount: DonationAmount
        The target amount to be raised during the campaign.
    started_at: str
        The timestamp when the charity campaign started.
    """
    id: str
    charity_name: str
    charity_description: str
    charity_logo: str
    charity_website: str
    current_amount: DonationAmount
    target_amount: DonationAmount
    started_at: str


class CharityCampaignProgressEvent(Broadcaster):
    """
    Represents the progress of an ongoing charity campaign during a Twitch stream.

    Attributes
    ----------
    id: str
        The unique identifier for the charity campaign.
    charity_name: str
        The name of the charity.
    charity_description: str
        A description of the charity.
    charity_logo: str
        The URL of the charity's logo.
    charity_website: str
        The URL of the charity's website.
    current_amount: DonationAmount
        The current amount raised during the campaign.
    target_amount: DonationAmount
        The target amount to be raised during the campaign.
    """
    id: str
    charity_name: str
    charity_description: str
    charity_logo: str
    charity_website: str
    current_amount: DonationAmount
    target_amount: DonationAmount


class CharityCampaignStopEvent(Broadcaster):
    """
    Represents the end of a charity campaign during a Twitch stream.

    Attributes
    ----------
    id: str
        The unique identifier for the charity campaign.
    charity_name: str
        The name of the charity.
    charity_description: str
        A description of the charity.
    charity_logo: str
        The URL of the charity's logo.
    charity_website: str
        The URL of the charity's website.
    current_amount: DonationAmount
        The current amount raised during the campaign.
    target_amount: DonationAmount
        The target amount to be raised during the campaign.
    stopped_at: str
        The timestamp when the charity campaign ended.
    """
    id: str
    charity_name: str
    charity_description: str
    charity_logo: str
    charity_website: str
    current_amount: DonationAmount
    target_amount: DonationAmount
    stopped_at: str


class GoalsEvent(SpecificBroadcaster):
    """
    Represents a goal event during a Twitch stream.

    Attributes
    ----------
    id: str
        The unique identifier for the goal event.
    type: str
        The type of goal being tracked. Possible values include:

        - follow
        - subscription
        - subscription_count
        - new_subscription
        - new_subscription_count
        - new_bit
        - new_cheerer
    description: str
        A description of the goal event.
    is_achieved: bool
        Indicates whether the goal has been achieved.
    current_amount: int
        The current progress towards the goal.
    target_amount: int
        The target amount to achieve the goal.
    started_at: str
        The timestamp when the goal event started.
    ended_at: str
        The timestamp when the goal event ended, if applicable.
    """
    id: str
    type: Literal[
        'follow',
        'subscription',
        'subscription_count',
        'new_subscription',
        'new_subscription_count',
        'new_bit',
        'new_cheerer'
    ]
    description: str
    is_achieved: bool
    current_amount: int
    target_amount: int
    started_at: str
    ended_at: str



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the 'Software'),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from enum import Enum

__all__ = ('Scopes',)


class Scopes(str, Enum):
    """
    An enumeration representing the available OAuth scopes.

    Each scope defines a specific permission that an application can request during
    the OAuth authentication process.

    Attributes
    ----------
    ANALYTICS_READ_EXTENSIONS : str
        Allows reading analytics data for extensions owned by the authenticated user.
    ANALYTICS_READ_GAMES : str
        Allows reading analytics data for games owned by the authenticated user.
    BITS_READ : str
        Allows reading information about Bits transactions involving the user.
    CHANNEL_BOT : str
        Grants access to manage and interact with a channel on behalf of the bot.
    CHANNEL_MANAGE_ADS : str
        Allows the application to start and manage ads for the authenticated user's channel.
    CHANNEL_READ_ADS : str
        Allows reading ad information for the authenticated user's channel.
    CHANNEL_MANAGE_BROADCAST : str
        Grants access to manage broadcasts, such as updating titles or categories.
    CHANNEL_READ_CHARITY : str
        Allows reading charity event details for the authenticated channel.
    CHANNEL_EDIT_COMMERCIAL : str
        Allows editing and running commercials on the channel.
    CHANNEL_READ_EDITORS : str
        Grants access to see the list of editors for a channel.
    CHANNEL_MANAGE_EXTENSIONS : str
        Allows management of extensions on the channel.
    CHANNEL_READ_GOALS : str
        Grants access to read goal information for the channel.
    CHANNEL_READ_GUEST_STAR : str
        Allows reading information about Guest Star events.
    CHANNEL_MANAGE_GUEST_STAR : str
        Allows managing Guest Star events for the channel.
    CHANNEL_READ_HYPE_TRAIN : str
        Grants access to read Hype Train events and statistics.
    CHANNEL_MANAGE_MODERATORS : str
        Allows management of moderators on the channel.
    CHANNEL_READ_POLLS : str
        Allows reading poll information for the channel.
    CHANNEL_MANAGE_POLLS : str
        Grants access to create and manage polls on the channel.
    CHANNEL_READ_PREDICTIONS : str
        Allows reading prediction data for the channel.
    CHANNEL_MANAGE_PREDICTIONS : str
        Allows managing predictions on the channel.
    CHANNEL_MANAGE_RAIDS : str
        Grants access to manage raids for the channel.
    CHANNEL_READ_REDEMPTIONS : str
        Allows reading channel point redemptions.
    CHANNEL_MANAGE_REDEMPTIONS : str
        Allows managing channel point redemptions.
    CHANNEL_MANAGE_SCHEDULE : str
        Grants access to manage the schedule of the channel.
    CHANNEL_READ_STREAM_KEY : str
        Allows reading the stream key of the channel.
    CHANNEL_READ_SUBSCRIPTIONS : str
        Allows reading subscription information for the channel.
    CHANNEL_MANAGE_VIDEOS : str
        Grants access to manage videos on the channel.
    CHANNEL_READ_VIPS : str
        Allows reading VIP information for the channel.
    CHANNEL_MANAGE_VIPS : str
        Allows managing VIPs on the channel.
    CLIPS_EDIT : str
        Grants access to edit clips for the channel.
    MODERATION_READ : str
        Allows reading moderation-related information for a channel.
    MODERATOR_MANAGE_ANNOUNCEMENTS : str
        Grants access to manage announcements on the channel.
    MODERATOR_MANAGE_AUTOMOD : str
        Allows managing AutoMod settings for the channel.
    MODERATOR_READ_AUTOMOD_SETTINGS : str
        Allows reading AutoMod settings for the channel.
    MODERATOR_MANAGE_AUTOMOD_SETTINGS : str
        Grants access to manage AutoMod settings for the channel.
    MODERATOR_MANAGE_BANNED_USERS : str
        Allows managing banned users for the channel.
    MODERATOR_READ_BLOCKED_TERMS : str
        Grants access to read blocked terms on the channel.
    MODERATOR_READ_CHAT_MESSAGES : str
        Allows reading chat messages on the channel.
    MODERATOR_MANAGE_BLOCKED_TERMS : str
        Grants access to manage blocked terms for the channel.
    MODERATOR_MANAGE_CHAT_MESSAGES : str
        Allows managing chat messages, including deleting them.
    MODERATOR_READ_CHAT_SETTINGS : str
        Allows reading chat settings for the channel.
    MODERATOR_MANAGE_CHAT_SETTINGS : str
        Grants access to manage chat settings for the channel.
    MODERATOR_READ_CHATTERS : str
        Allows reading the list of chatters in the channel.
    MODERATOR_READ_FOLLOWERS : str
        Grants access to read followers' information on the channel.
    MODERATOR_READ_GUEST_STAR : str
        Allows reading information about Guest Star events.
    MODERATOR_MANAGE_GUEST_STAR : str
        Grants access to manage Guest Star events on the channel.
    MODERATOR_READ_SHIELD_MODE : str
        Allows reading shield mode settings for the channel.
    MODERATOR_MANAGE_SHIELD_MODE : str
        Grants access to manage shield mode settings.
    MODERATOR_READ_SHOUTOUTS : str
        Allows reading shoutout information for the channel.
    MODERATOR_MANAGE_SHOUTOUTS : str
        Grants access to manage shoutouts for the channel.
    MODERATOR_READ_SUSPICIOUS_USERS : str
        Allows reading information about suspicious users in the channel.
    MODERATOR_READ_UNBAN_REQUESTS : str
        Grants access to read unban requests for the channel.
    MODERATOR_MANAGE_UNBAN_REQUESTS : str
        Grants access to manage unban requests for the channel.
    MODERATOR_READ_WARNINGS : str
        Allows reading warning information for the channel.
    MODERATOR_MANAGE_WARNINGS : str
        Grants access to manage warnings for the channel.
    USER_BOT : str
        Grants access to interact with the user's account on behalf of a bot.
    USER_EDIT : str
        Allows editing user details, such as display name or bio.
    USER_EDIT_BROADCAST : str
        Grants access to edit broadcast information for the user.
    USER_READ_BLOCKED_USERS : str
        Allows reading the list of blocked users.
    USER_MANAGE_BLOCKED_USERS : str
        Grants access to manage blocked users.
    USER_READ_BROADCAST : str
        Allows reading broadcast information for the user.
    USER_READ_CHAT : str
        Allows reading chat messages sent by the user.
    USER_MANAGE_CHAT_COLOR : str
        Grants access to manage the chat color for the user.
    USER_READ_EMAIL : str
        Allows reading the user's email address.
    USER_READ_EMOTES : str
        Allows reading the list of emotes for the user.
    USER_READ_FOLLOWS : str
        Grants access to read the list of users followed by the user.
    USER_READ_MODERATED_CHANNELS : str
        Allows reading information about channels moderated by the user.
    USER_READ_SUBSCRIPTIONS : str
        Allows reading subscription information for the user.
    USER_READ_WHISPERS : str
        Grants access to read whispers sent and received by the user.
    USER_MANAGE_WHISPERS : str
        Grants access to manage whispers for the user.
    USER_WRITE_CHAT : str
        Allows sending chat messages on behalf of the user.
    """

    # Analytics
    ANALYTICS_READ_EXTENSIONS = 'analytics:read:extensions'
    ANALYTICS_READ_GAMES = 'analytics:read:games'

    # Bits
    BITS_READ = 'bits:read'

    # Channel
    CHANNEL_BOT = 'channel:bot'
    CHANNEL_MANAGE_ADS = 'channel:manage:ads'
    CHANNEL_READ_ADS = 'channel:read:ads'
    CHANNEL_MANAGE_BROADCAST = 'channel:manage:broadcast'
    CHANNEL_READ_CHARITY = 'channel:read:charity'
    CHANNEL_EDIT_COMMERCIAL = 'channel:edit:commercial'
    CHANNEL_READ_EDITORS = 'channel:read:editors'
    CHANNEL_MANAGE_EXTENSIONS = 'channel:manage:extensions'
    CHANNEL_READ_GOALS = 'channel:read:goals'
    CHANNEL_READ_GUEST_STAR = 'channel:read:guest_star'
    CHANNEL_MANAGE_GUEST_STAR = 'channel:manage:guest_star'
    CHANNEL_READ_HYPE_TRAIN = 'channel:read:hype_train'
    CHANNEL_MANAGE_MODERATORS = 'channel:manage:moderators'
    CHANNEL_READ_POLLS = 'channel:read:polls'
    CHANNEL_MANAGE_POLLS = 'channel:manage:polls'
    CHANNEL_READ_PREDICTIONS = 'channel:read:predictions'
    CHANNEL_MANAGE_PREDICTIONS = 'channel:manage:predictions'
    CHANNEL_MANAGE_RAIDS = 'channel:manage:raids'
    CHANNEL_READ_REDEMPTIONS = 'channel:read:redemptions'
    CHANNEL_MANAGE_REDEMPTIONS = 'channel:manage:redemptions'
    CHANNEL_MANAGE_SCHEDULE = 'channel:manage:schedule'
    CHANNEL_READ_STREAM_KEY = 'channel:read:stream_key'
    CHANNEL_READ_SUBSCRIPTIONS = 'channel:read:subscriptions'
    CHANNEL_MANAGE_VIDEOS = 'channel:manage:videos'
    CHANNEL_READ_VIPS = 'channel:read:vips'
    CHANNEL_MANAGE_VIPS = 'channel:manage:vips'
    CLIPS_EDIT = 'clips:edit'

    # Moderation
    MODERATION_READ = 'moderation:read'
    MODERATOR_MANAGE_ANNOUNCEMENTS = 'moderator:manage:announcements'
    MODERATOR_MANAGE_AUTOMOD = 'moderator:manage:automod'
    MODERATOR_READ_AUTOMOD_SETTINGS = 'moderator:read:automod_settings'
    MODERATOR_MANAGE_AUTOMOD_SETTINGS = 'moderator:manage:automod_settings'
    MODERATOR_MANAGE_BANNED_USERS = 'moderator:manage:banned_users'
    MODERATOR_READ_BLOCKED_TERMS = 'moderator:read:blocked_terms'
    MODERATOR_READ_CHAT_MESSAGES = 'moderator:read:chat_messages'
    MODERATOR_MANAGE_BLOCKED_TERMS = 'moderator:manage:blocked_terms'
    MODERATOR_MANAGE_CHAT_MESSAGES = 'moderator:manage:chat_messages'
    MODERATOR_READ_CHAT_SETTINGS = 'moderator:read:chat_settings'
    MODERATOR_MANAGE_CHAT_SETTINGS = 'moderator:manage:chat_settings'
    MODERATOR_READ_CHATTERS = 'moderator:read:chatters'
    MODERATOR_READ_FOLLOWERS = 'moderator:read:followers'
    MODERATOR_READ_GUEST_STAR = 'moderator:read:guest_star'
    MODERATOR_MANAGE_GUEST_STAR = 'moderator:manage:guest_star'
    MODERATOR_READ_SHIELD_MODE = 'moderator:read:shield_mode'
    MODERATOR_MANAGE_SHIELD_MODE = 'moderator:manage:shield_mode'
    MODERATOR_READ_SHOUTOUTS = 'moderator:read:shoutouts'
    MODERATOR_MANAGE_SHOUTOUTS = 'moderator:manage:shoutouts'
    MODERATOR_READ_SUSPICIOUS_USERS = 'moderator:read:suspicious_users'
    MODERATOR_READ_UNBAN_REQUESTS = 'moderator:read:unban_requests'
    MODERATOR_MANAGE_UNBAN_REQUESTS = 'moderator:manage:unban_requests'
    MODERATOR_READ_WARNINGS = 'moderator:read:warnings'
    MODERATOR_MANAGE_WARNINGS = 'moderator:manage:warnings'

    # User
    USER_BOT = 'user:bot'
    USER_EDIT = 'user:edit'
    USER_EDIT_BROADCAST = 'user:edit:broadcast'
    USER_READ_BLOCKED_USERS = 'user:read:blocked_users'
    USER_MANAGE_BLOCKED_USERS = 'user:manage:blocked_users'
    USER_READ_BROADCAST = 'user:read:broadcast'
    USER_READ_CHAT = 'user:read:chat'
    USER_MANAGE_CHAT_COLOR = 'user:manage:chat_color'
    USER_READ_EMAIL = 'user:read:email'
    USER_READ_EMOTES = 'user:read:emotes'
    USER_READ_FOLLOWS = 'user:read:follows'
    USER_READ_MODERATED_CHANNELS = 'user:read:moderated_channels'
    USER_READ_SUBSCRIPTIONS = 'user:read:subscriptions'
    USER_READ_WHISPERS = 'user:read:whispers'
    USER_MANAGE_WHISPERS = 'user:manage:whispers'
    USER_WRITE_CHAT = 'user:write:chat'



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Dict, Optional, Any, Union
    from aiohttp import ClientWebSocketResponse
    from aiohttp import ClientResponse


__all__ = ('TwitchException',
           'ClientException',
           'HTTPException',
           'TwitchServerError',
           'Forbidden',
           'NotFound',
           'AuthFailure',
           'ConnectionClosed',
           'UnregisteredUser')


class TwitchException(Exception):
    """
    Base exception for twitch.py.
    """
    pass


class ClientException(TwitchException):
    """
    Exception raised when an operation in the Client class fails.
    """
    pass


class HTTPException(TwitchException):
    """Exception raised for failed HTTP requests."""

    def __init__(self, response: ClientResponse, message: Optional[Union[str, Dict[str, Any]]]):
        self.response = response
        self.status = response.status if hasattr(response, 'status') else 0

        if isinstance(message, dict):
            self.code = message.get('status', 0)
            self.text = message.get('message', '')
        else:
            self.text = message or ''
            self.code = 0
        super().__init__(f'{self.response.reason} (error code: {self.code}): {self.text}')


class TwitchServerError(HTTPException):
    """Exception raised when a 500 range status code occurs."""
    pass


class Forbidden(HTTPException):
    """Exception raised when status code 403 occurs."""
    pass


class NotFound(HTTPException):
    """Exception raised when status code 404 occurs."""
    pass


class AuthFailure(ClientException):
    """Exception raised when authentication fails, typically due to invalid credentials."""
    pass


class UnregisteredUser(ClientException):
    """Exception raised when the user is UnregisteredUse."""
    pass


class ConnectionClosed(ClientException):
    """
    Exception raised when the WebSocket connection is closed unexpectedly.
    """

    def __init__(self, socket: ClientWebSocketResponse, *, code: Optional[int] = None):
        self.code: int = code or socket.close_code or -1
        self.reason: str = ''
        super().__init__(f'WebSocket closed with {self.code} close code.')



"""
twitch.ext.oauth

An extension module to facilitate the authorization

:copyright: (c) 2024-present Snifo
:license: MIT, see LICENSE for more details.
"""

from .oauth import DeviceAuthFlow
from .enums import Scopes



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from .users import SpecificBroadcaster, SpecificModerator
from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from typing import Literal


# Ad
class AdBreakBeginEvent(SpecificBroadcaster):
    """
    Represents an event where an ad break begins.

    Attributes
    ----------
    duration_seconds: int
        The duration of the ad break in seconds.
    started_at: str
        The timestamp when the ad break started, in ISO 8601 format.
    is_automatic: bool
        Whether the ad break was triggered automatically.
    requester_user_id: str
        The user ID of the requester who initiated the ad break.
    requester_user_login: str
        The login name of the requester.
    requester_user_name: str
        The display name of the requester.
    """
    duration_seconds: int
    started_at: str
    is_automatic: bool
    requester_user_id: str
    requester_user_login: str
    requester_user_name: str


# Raid
class RaidEvent(TypedDict):
    """
    Represents an event where a raid occurs.

    Attributes
    ----------
    from_broadcaster_user_id: str
        The user ID of the broadcaster initiating the raid.
    from_broadcaster_user_login: str
        The login name of the broadcaster initiating the raid.
    from_broadcaster_user_name: str
        The display name of the broadcaster initiating the raid.
    to_broadcaster_user_id: str
        The user ID of the broadcaster receiving the raid.
    to_broadcaster_user_login: str
        The login name of the broadcaster receiving the raid.
    to_broadcaster_user_name: str
        The display name of the broadcaster receiving the raid.
    viewers: int
        The number of viewers participating in the raid.
    """
    from_broadcaster_user_id: str
    from_broadcaster_user_login: str
    from_broadcaster_user_name: str
    to_broadcaster_user_id: str
    to_broadcaster_user_login: str
    to_broadcaster_user_name: str
    viewers: int


# Shoutout
class ShoutoutCreateEvent(SpecificBroadcaster, SpecificModerator):
    """
    Represents an event where a shoutout is created.

    Attributes
    ----------
    to_broadcaster_user_id: str
        The user ID of the broadcaster receiving the shoutout.
    to_broadcaster_user_login: str
        The login name of the broadcaster receiving the shoutout.
    to_broadcaster_user_name: str
        The display name of the broadcaster receiving the shoutout.
    viewer_count: int
        The number of viewers when the shoutout was created.
    started_at: str
        The timestamp when the shoutout started, in ISO 8601 format.
    cooldown_ends_at: str
        The timestamp when the cooldown period for the shoutout ends, in ISO 8601 format.
    target_cooldown_ends_at: str
        The timestamp when the target cooldown period ends, in ISO 8601 format.
    """
    to_broadcaster_user_id: str
    to_broadcaster_user_login: str
    to_broadcaster_user_name: str
    viewer_count: int
    started_at: str
    cooldown_ends_at: str
    target_cooldown_ends_at: str


class ShoutoutReceivedEvent(SpecificBroadcaster):
    """
    Represents an event where a shoutout is received.

    Attributes
    ----------
    from_broadcaster_user_id: str
        The user ID of the broadcaster who sent the shoutout.
    from_broadcaster_user_login: str
        The login name of the broadcaster who sent the shoutout.
    from_broadcaster_user_name: str
        The display name of the broadcaster who sent the shoutout.
    viewer_count: int
        The number of viewers when the shoutout was received.
    started_at: str
        The timestamp when the shoutout was received, in ISO 8601 format.
    """
    from_broadcaster_user_id: str
    from_broadcaster_user_login: str
    from_broadcaster_user_name: str
    viewer_count: int
    started_at: str


# Stream
class StreamOnlineEvent(SpecificBroadcaster):
    """
    Represents an event where a stream goes online.

    Attributes
    ----------
    id: str
        The ID of the stream.
    type: Literal['live', 'playlist', 'watch_party', 'premiere', 'rerun']
        The type of the stream.
    started_at: s²tr
        The timestamp when the stream started, in ISO 8601 format.
    """
    id: str
    type: Literal['live', 'playlist', 'watch_party', 'premiere', 'rerun']
    started_at: str


class StreamOfflineEvent(SpecificBroadcaster):
    """
    Represents an event where a stream goes offline.

    Attributes
    ----------
    (No additional attributes; inherits from SpecificBroadcaster)
    """
    pass



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from typing import NotRequired, Optional, List


class Broadcaster(TypedDict):
    """
    Represents a broadcaster with basic information.

    Attributes
    ----------
    broadcaster_id: str
        The ID of the broadcaster.
    broadcaster_name: str
        The display name of the broadcaster.
    broadcaster_login: str
        The login name of the broadcaster.
    """
    broadcaster_id: str
    broadcaster_name: str
    broadcaster_login: str


class SpecificBroadcaster(TypedDict):
    """
    Represents a specific broadcaster with user details.

    Attributes
    ----------
    broadcaster_user_id: str
        The user ID of the broadcaster.
    broadcaster_user_name: str
        The user name of the broadcaster.
    broadcaster_user_login: str
        The login name of the broadcaster.
    """
    broadcaster_user_id: str
    broadcaster_user_name: str
    broadcaster_user_login: str


class Moderator(TypedDict):
    """
    Represents a moderator with basic information.

    Attributes
    ----------
    moderator_id: str
        The ID of the moderator.
    moderator_name: str
        The display name of the moderator.
    moderator_login: str
        The login name of the moderator.
    """
    moderator_id: str
    moderator_name: str
    moderator_login: str


class SpecificModerator(TypedDict):
    """
    Represents a specific moderator with user details.

    Attributes
    ----------
    moderator_user_id: str
        The user ID of the moderator.
    moderator_user_name: str
        The user name of the moderator.
    moderator_user_login: str
        The login name of the moderator.
    """
    moderator_user_id: str
    moderator_user_name: str
    moderator_user_login: str


class SpecificUser(TypedDict):
    """
    Represents a specific user with details.

    Attributes
    ----------
    user_id: str
        The ID of the user.
    user_name: str
        The name of the user.
    user_login: str
        The login name of the user.
    """
    user_id: str
    user_name: str
    user_login: str


# User
class UserUpdateEvent(TypedDict, total=False):
    """
    Represents an update event for a user.

    Attributes
    ----------
    user_id: str
        The ID of the user.
    user_login: str
        The login name of the user.
    user_name: str
        The name of the user.
    email: str, optional
        The email of the user, if provided.
    email_verified: bool, optional
        Whether the email has been verified.
    description: str
        The description or bio of the user.
    """
    user_id: str
    user_login: str
    user_name: str
    email: NotRequired[str]
    email_verified: NotRequired[bool]
    description: str


# Whisper
class WhisperDetails(TypedDict):
    """
    Represents the details of a whisper message.

    Attributes
    ----------
    text: str
        The content of the whisper message.
    """
    text: str


class WhisperReceivedEvent(TypedDict):
    """
    Represents an event where a whisper message is received.

    Attributes
    ----------
    from_user_id: str
        The ID of the user who sent the whisper.
    from_user_name: str
        The name of the user who sent the whisper.
    from_user_login: str
        The login name of the user who sent the whisper.
    to_user_id: str
        The ID of the user who received the whisper.
    to_user_name: str
        The name of the user who received the whisper.
    to_user_login: str
        The login name of the user who received the whisper.
    whisper_id: str
        The ID of the whisper message.
    whisper: WhisperDetails
        The details of the whisper message.
    """
    from_user_id: str
    from_user_name: str
    from_user_login: str
    to_user_id: str
    to_user_name: str
    to_user_login: str
    whisper_id: str
    whisper: WhisperDetails


# Automod
class Cheermote(TypedDict):
    """
    Represents a cheermote with its properties.

    Attributes
    ----------
    prefix: str
        The prefix of the cheermote.
    bits: int
        The number of bits required to use the cheermote.
    tier: int
        The tier level of the cheermote.
    """
    prefix: str
    bits: int
    tier: int


class Emote(TypedDict):
    """
    Represents an emote with its properties.

    Attributes
    ----------
    id: str
        The ID of the emote.
    emote_set_id: str
        The ID of the emote set to which the emote belongs.
    """
    id: str
    emote_set_id: str


class Fragment(TypedDict):
    """
    Represents a fragment of a message that can be a cheermote or emote.

    Attributes
    ----------
    cheermote: Cheermote, optional
        The cheermote fragment, if any.
    emote: Emote, optional
        The emote fragment, if any.
    """
    cheermote: Optional[Cheermote]
    emote: Optional[Emote]


class Message(TypedDict):
    """
    Represents a message with its content and fragments.

    Attributes
    ----------
    text: str
        The content of the message.
    fragments: List[Fragment]
        A list of fragments in the message.
    """
    text: str
    fragments: List[Fragment]


class MessageHoldEvent(SpecificBroadcaster, SpecificUser):
    """
    Represents an event where a message is held for moderation.

    Attributes
    ----------
    message_id: str
        The ID of the held message.
    message: Message
        The details of the held message.
    """
    message_id: str
    message: Message


class MessageUpdateEvent(SpecificBroadcaster, SpecificUser):
    """
    Represents an event where a message is updated.

    Attributes
    ----------
    status: str
        The new status of the message.
    message_id: str
        The ID of the updated message.
    message: Message
        The details of the updated message.
    """
    status: str
    message_id: str
    message: Message


# Auth
class AuthorizationGrantEvent(SpecificUser):
    """
    Represents an event where authorization is granted.

    Attributes
    ----------
    client_id: str
        The ID of the client that received authorization.
    """
    client_id: str


class AuthorizationRevokeEvent(TypedDict):
    """
    Represents an event where authorization is revoked.

    Attributes
    ----------
    client_id: str
        The ID of the client whose authorization was revoked.
    user_id: str
        The ID of the user whose authorization was revoked.
    user_login: str, optional
        The login name of the user, if available.
    user_name: str, optional
        The name of the user, if available.
    """
    client_id: str
    user_id: str
    user_login: Optional[str]
    user_name: Optional[str]



"""
The MIT License (MIT)

Copyright (c) 2024-present Snifo

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT firstED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from twitch import __version__, __github__, errors, utils
from typing import TYPE_CHECKING
import aiohttp
import asyncio
import socket
import time

if TYPE_CHECKING:
    from typing import Any, List, Optional, ClassVar
    from twitch.types import users

import logging
_logger = logging.getLogger(__name__)

__all__ = ('HTTPClient',)


class HTTPClient:
    """Represents an asynchronous HTTP client for sending HTTP requests."""

    OAUTH2: ClassVar[str] = 'https://id.twitch.tv/oauth2/'

    def __init__(self, client_id: str, *, loop: Optional[asyncio.AbstractEventLoop] = None) -> None:
        self.__session: Optional[aiohttp.ClientSession] = None
        self.user_agent: str = f'twitch.py/{__version__} (GitHub: {__github__})'
        self.client_id: str = client_id
        self.loop: asyncio.AbstractEventLoop = loop

    async def close(self) -> None:
        """
        Closes the underlying session if it exists.
        """
        if self.__session is not None and not self.__session.closed:
            await self.__session.close()

    async def clear_session(self) -> None:
        """
        Clears the session and cancels any pending keep-alive tasks.
        """
        if self.__session and self.__session.closed:
            self.__session = None

    async def request(self, method: str, path: str, **kwargs: Any) -> Any:
        """
        Make an HTTP request with the provided route and keyword arguments.
        """
        if self.__session is None:
            connector = aiohttp.TCPConnector(limit=0, family=socket.AF_INET)
            self.__session = aiohttp.ClientSession(connector=connector)

        url = self.OAUTH2 + path
        kwargs['headers'] = {'Client-ID': self.client_id}

        for attempt in range(3):
            try:
                async with self.__session.request(method, url, **kwargs) as response:
                    _logger.debug('%s >> %s with %s has returned status code %s', method, url,
                                  kwargs.get('data'),
                                  response.status)
                    data = await utils.json_or_text(response)
                    if 300 > response.status >= 200:
                        _logger.debug('%s << %s has received %s', method, url, data)
                        return data
                    if response.status == 403:
                        raise errors.Forbidden(response, data)
                    elif response.status >= 500:
                        raise errors.TwitchServerError(response, data)
                    else:
                        raise errors.HTTPException(response, data)
            except OSError as e:
                if attempt < 2 and e.errno in (54, 10054):
                    await asyncio.sleep(1 + attempt * 2)
                    continue
                raise

    async def start_device_auth_flow(self, scopes: List[str]) -> users.DeviceAuthFlow:
        """
        Starts the device authorization flow.
        """
        scopes = list(dict.fromkeys(scopes))
        body = {
            'client_id': self.client_id,
            'scopes': ' '.join(scopes)
        }
        data = await self.request('POST', 'device', data=body)
        return data

    async def poll_for_token(self, device_code: str, expires_in: int, interval: int) -> users.OAuthRefreshToken:
        """
        Polls the Twitch API to check if the user has authorized the application.
        """
        start_time = time.time()
        while time.time() - start_time < (expires_in - 15):
            await asyncio.sleep(interval)
            try:
                data = await self.check_device_code(device_code=device_code)
                _logger.debug('Successfully authorized using device code flow.')
                return data
            except errors.HTTPException as error:
                if error.text in ['authorization_pending', 'slow_down']:
                    continue
                raise
        raise TimeoutError('Authorization expired.')

    async def check_device_code(self, device_code: str) -> Any:
        """
        Checks the device code to obtain the access token.
        """
        body = {
            'client_id': self.client_id,
            'device_code': device_code,
            'grant_type': 'urn:ietf:params:oauth:grant-type:device_code'
        }
        data = await self.request('POST', 'token', data=body,
                                  headers={'Content-Type': 'application/x-www-form-urlencoded'})
        return data

    async def revoke_token(self, token: str) -> None:
        """
        Revokes a token.
        """
        body = {
            'client_id': self.client_id,
            'token': token,
        }
        await self.request('POST', 'revoke', data=body)

